#include <iostream>
#include <fstream>
#include <string>
#include <memory.h>
#include <cstdio>
#include <vector>

#include "clb/clbLib.h"
#include "clb/util/GeometryFactory.h"
#include "clb/util/Permutation.h"

using namespace std;
using namespace clb;
using namespace clb::Global;
using namespace clb::math;

// No 3D window, for searching the solution on console ui.
#define CONSOLE_APP

#ifdef CONSOLE_APP
#define PRINT(...) { printf(__VA_ARGS__); printf("\n"); }
#else
#define PRINT(...) clb::Global::console(__VA_ARGS__)
#endif

/// Defines a single 2x2x8 piece of the puzzle.
struct Block
{
	int o[2*2*8]; // x 2,y 2,z 8

	void Clear() { memset(this, 0, sizeof(*this)); }

	int Get(int x, int y, int z) const { return o[z+y*8+x*16]; }
	void Set(int x, int y, int z, int v)
	{ 
		assert(x >= 0);
		assert(y >= 0);
		assert(z >= 0);
		assert(x < 2);
		assert(y < 2);
		assert(z < 8);
		o[z+y*8+x*16] = v;
	}

	bool operator==(const Block &rhs) const
	{
		for(int i = 0; i < 2*2*8; ++i)
			if (o[i] != rhs.o[i])
				return false;
		return true;
	}
};

/// A collection of all rotations of a single piece, with symmetries removed. These are cached for efficiency.
struct PreRotatedBlock
{
	std::vector<Block> blocks;
};

/// Problem configuration invariants
const int nBigBlocks = 12;
const int nSmallBlocks = 6;
const int nBlocks = 18;
PreRotatedBlock prBig[nBigBlocks];
PreRotatedBlock prSmall[nSmallBlocks];

// Input data for the big blocks - each element is of the form
//{ [x=0,y=0,z=0-7], [x=0,y=1,z=0-7],
//  [x=1,y=0,z=0-7], [x=1,y=1,z=0-7], },
const Block cBigBlockInput[nBigBlocks] =
{
{ 1,1,1,1,1,1,1,1,  1,1,0,0,0,0,0,0,
  1,1,0,0,0,0,1,1,  1,1,0,0,0,0,0,0, },

{ 1,1,1,1,1,1,1,1,  1,1,0,0,0,0,1,1,
  1,1,0,1,1,0,1,1,  1,1,0,0,0,0,1,1, },

{ 1,1,1,1,1,1,1,1,  1,1,0,0,0,0,1,1,
  1,1,0,0,0,0,1,1,  1,1,0,0,0,0,1,1, },

{ 1,1,1,1,1,1,1,1,  1,0,0,0,0,0,0,1,
  1,1,0,0,0,0,1,1,  1,0,0,0,0,0,0,1, },

{ 1,1,1,1,1,1,1,1,  1,1,0,0,0,0,1,1,
  0,0,0,1,1,0,0,0,  0,0,0,0,0,0,0,0, },

{ 1,1,1,1,1,1,1,1,  0,0,0,0,0,0,0,0, 
  1,1,0,0,0,1,1,1,  0,0,0,0,0,0,0,0, },

{ 1,1,1,1,1,1,1,1,  0,0,0,0,0,0,0,1,
  1,1,1,0,0,1,1,1,  0,0,0,0,0,0,0,1, },

{ 1,1,1,1,1,1,1,1,  1,1,1,0,0,1,1,1,
  0,0,0,1,1,0,0,0,  0,0,0,0,0,0,0,0, },

{ 1,1,1,1,1,1,1,1,  1,1,1,0,0,1,1,1,
  0,0,0,0,0,0,0,1,  0,0,0,0,0,0,0,1, },

{ 1,1,1,1,1,1,1,1,  1,1,0,0,0,0,1,1,
  0,0,0,0,0,0,0,0,  0,0,0,0,0,0,0,0, },

{ 1,1,1,1,1,1,1,1,  0,0,0,0,0,0,1,1, 
  1,1,1,0,0,0,1,1,  0,0,0,0,0,0,1,1, },

{ 1,1,1,1,1,1,1,1, 1,1,0,0,0,0,1,1,
  0,0,0,1,1,0,1,1, 0,0,0,0,0,0,1,1, },

};

// Input data for the small blocks - each element is of the form
//{ [x=0,y=0,z=0-7], [x=0,y=1,z=0-7],
//  [x=1,y=0,z=0-7], [x=1,y=1,z=0-7], },
const Block cSmallBlockInput[nSmallBlocks] =
{

{ 1,1,1,1,1,1,1,1,  1,1,1,1,1,1,1,1,
  1,1,1,1,1,1,1,1,  1,1,1,1,1,1,1,1, },

{ 1,1,1,1,1,1,1,1,  1,1,1,1,1,1,1,1,
  1,1,0,0,0,0,1,1,  1,1,0,0,0,0,1,1, },

{ 1,1,1,1,1,1,1,1,  1,1,1,1,1,1,1,1,
  1,1,1,1,0,0,1,1,  1,1,1,1,0,0,1,1, },

{ 1,1,1,1,1,1,1,1,  1,1,1,1,0,0,1,1,
  1,1,1,1,0,0,1,1,  1,1,1,1,0,0,1,1, },

{ 1,1,1,1,1,1,1,1,  1,1,0,0,1,1,1,1,
  1,1,0,0,0,0,1,1,  1,1,0,0,0,0,1,1, },

{ 1,1,1,1,1,1,1,1,  1,1,1,1,0,0,1,1,
  1,1,0,0,1,1,1,1,  1,1,0,0,0,0,1,1, },

};

// Possible positions for big blocks, [x,y,z]
const int cBigBlockPos[] =
{ 
	2, 2, 0,
	4, 2, 0,
	2, 4, 0,
	4, 4, 0,

	0, 2, 2,
	0, 4, 2,
	0, 2, 4,
	0, 4, 4,

	2, 0, 2,
	4, 0, 2,
	2, 0, 4,
	4, 0, 4,

	2, 2, 7,
	4, 2, 7,
	2, 4, 7,
	4, 4, 7,

	7, 2, 2,
	7, 4, 2,
	7, 2, 4,
	7, 4, 4,

	2, 7, 2,
	4, 7, 2,
	2, 7, 4,
	4, 7, 4,

};

// After applying rotation symmetry reduction, there are two possible parities for the solution
// w.r.t. to small and big blocks.

// Possible positions for small blocks, [x,y,z]
// 1st parity configuration
const int cSmallBlockPos[] =
{ 
	// Along Z axis
	3, 0, 0,
	3, 6, 0,

	// Along X axis
	0, 3, 0,
	0, 3, 6,

	// Along Y axis
	0, 0, 3,
	6, 0, 3,

	// Along neg. Z axis
	3, 0, 7,
	3, 6, 7,

	// Along neg. X axis
	7, 3, 0,
	7, 3, 6,

	// Along neg. Y axis
	0, 7, 3,
	6, 7, 3,
};

// Possible positions for small blocks, [x,y,z]
// 2nd parity for small blocks
const int cSmallBlockPos2[] =
{
	// Along Z axis
	0, 3, 0,
	6, 3, 0,

	// Along X axis
	0, 0, 3,
	0, 6, 3,

	// Along Y axis
	3, 0, 0,
	3, 0, 6,

	// Along neg. Z axis
	0, 3, 7,
	6, 3, 7,

	// Along neg. X axis
	7, 0, 3,
	7, 6, 3,

	// Along neg. Y axis
	3, 7, 0,
	3, 7, 6,
};

// 6 3x3 rotation matrices concatenated together, used to determine which way a block with given orientation index is orientated.
const int cBigBlockAxisIncr[6*9] =
{
	1,0,0, 0,1,0, 0,0,1,
	0,0,1, 1,0,0, 0,1,0,
	0,1,0, 0,0,1, 1,0,0,

	1,0,0, 0,1,0, 0,0,-1,
	0,0,-1, 1,0,0, 0,1,0,
	0,1,0, 0,0,-1, 1,0,0,
};

/// Generates a block that is rotated 90 degrees about the Z axis
void RotateBlockZ90(Block &dst, const Block &b)
{
	for(int y = 0; y < 2; ++y)
		for(int x = 0; x < 2; ++x)
			for(int z = 0; z < 8; ++z)
				dst.Set(y, 1-x, z, b.Get(x, y, z));
}

/// Generates a block that is rotated 180 degrees about the Y axis
void RotateBlockY180(Block &dst, const Block &b)
{
	for(int y = 0; y < 2; ++y)
		for(int x = 0; x < 2; ++x)
			for(int z = 0; z < 8; ++z)
				dst.Set(1-x, y, 7-z, b.Get(x, y, z));
}

/// Generates and saves all rotations of a piece and removes symmetries.
void GeneratePreRotatedBlock(PreRotatedBlock &dst, const Block &b)
{
	dst.blocks.push_back(b); // config 0
	for(int j = 0; j < 2; ++j)
	{
		for(int i = 0; i < 3; ++i) // configs 1-3 (when j==0) and configs 5-7 (when j==1)
		{
			Block rot;
			RotateBlockZ90(rot, dst.blocks.back());
			dst.blocks.push_back(rot);
		}
		if (j == 0) // config 4
		{
			Block rot;
			RotateBlockY180(rot, dst.blocks.back());
			dst.blocks.push_back(rot);
		}
	}

	// prune duplicates
	for(int i = 0; i < dst.blocks.size()-1; ++i)
		for(int j = i+1; j < dst.blocks.size(); ++j)
			if (dst.blocks[i] == dst.blocks[j])
			{
				dst.blocks.erase(dst.blocks.begin() + j);
				--j;
			}
	
	PRINT("%d configurations.", (int)dst.blocks.size());
	assert(dst.blocks.size() > 0);
	assert(dst.blocks.size() <= 8);
}

void PrecalculateRotations()
{
	for(int i = 0; i < nBigBlocks; ++i)
		GeneratePreRotatedBlock(prBig[i], cBigBlockInput[i]);
	for(int i = 0; i < nSmallBlocks; ++i)
		GeneratePreRotatedBlock(prSmall[i], cSmallBlockInput[i]);
}

typedef unsigned int BlockID;

// Empty positions in the block grid have this value, others hold the index to the block that occupies the space.
static const BlockID cEmptyGridPos = 0xFFFFFFFF;

struct BlockGrid
{
public:
	static const int cGridSize = 8;

	void Clear() { memset(this, cEmptyGridPos, sizeof(*this)); }

	BlockID Get(int x, int y, int z) const
	{
		assert(x >= 0);
		assert(y >= 0);
		assert(z >= 0);
		assert(x < cGridSize);
		assert(y < cGridSize);
		assert(z < cGridSize);
		return o[z][y][x];
	}
	void Set(int x, int y, int z, BlockID id)
	{
		assert(Get(x,y,z) || true);
		o[z][y][x] = id;
	}

	bool IsEmpty(int x, int y, int z) const { return Get(x,y,z) == cEmptyGridPos; }

	/// @param pos [0,#elements in blockPositions-1] The index of the position to use
	/// @param smallBlock If true, placing a small block instead of a big block.
	/// @param id The id to fill the grid with.
	/// @return Returns true if the block doesn't intersect any of the existing objects.
	/// If returns true, the block was added to the grid. If false, no cell of the grid were modified.
	bool TryPlaceBlock(const Block &b, int pos, bool smallBlock, int id)
	{
		const int *blockPositions = smallBlock ? cSmallBlockPos2 : cBigBlockPos;
		int startX = blockPositions[pos*3];
		int startY = blockPositions[pos*3+1];
		int startZ = blockPositions[pos*3+2];

		// block orientation direction, the divisor comes from the num. of positions in one direction
		int i = smallBlock ? (pos/2) : (pos/4);

		for(int z = 0; z < 8; ++z)
			for(int y = 0; y < 2; ++y)
				for(int x = 0; x < 2; ++x)
				{
					int X = startX + x * cBigBlockAxisIncr[i*9] + y * cBigBlockAxisIncr[i*9+1] + z * cBigBlockAxisIncr[i*9+2];
					int Y = startY + x * cBigBlockAxisIncr[i*9+3] + y * cBigBlockAxisIncr[i*9+4] + z * cBigBlockAxisIncr[i*9+5];
					int Z = startZ + x * cBigBlockAxisIncr[i*9+6] + y * cBigBlockAxisIncr[i*9+7] + z * cBigBlockAxisIncr[i*9+8];

					if (b.Get(x,y,z) && !IsEmpty(X,Y,Z))
						return false;
				}

		for(int z = 0; z < 8; ++z)
			for(int y = 0; y < 2; ++y)
				for(int x = 0; x < 2; ++x)
				{
					int X = startX + x * cBigBlockAxisIncr[i*9] + y * cBigBlockAxisIncr[i*9+1] + z * cBigBlockAxisIncr[i*9+2];
					int Y = startY + x * cBigBlockAxisIncr[i*9+3] + y * cBigBlockAxisIncr[i*9+4] + z * cBigBlockAxisIncr[i*9+5];
					int Z = startZ + x * cBigBlockAxisIncr[i*9+6] + y * cBigBlockAxisIncr[i*9+7] + z * cBigBlockAxisIncr[i*9+8];

					if (b.Get(x,y,z))
						o[Z][Y][X] = id; 
				}

		return true;
	}

	// Frees the space occupied by a block in the grid.
	/// @param pos [0,#elements in blockPositions-1] The index of the position to use
	/// @param smallBlock If true, placing a small block instead of a big block.
	void ClearBlock(const Block &b, int pos, bool smallBlock)
	{
		const int *blockPositions = smallBlock ? cSmallBlockPos2 : cBigBlockPos;
		int startX = blockPositions[pos*3];
		int startY = blockPositions[pos*3+1];
		int startZ = blockPositions[pos*3+2];

		// block orientation direction, the divisor comes from the num. of positions in one direction
		int i = smallBlock ? (pos/2) : (pos/4);

		for(int z = 0; z < 8; ++z)
			for(int y = 0; y < 2; ++y)
				for(int x = 0; x < 2; ++x)
				{
					int X = startX + x * cBigBlockAxisIncr[i*9] + y * cBigBlockAxisIncr[i*9+1] + z * cBigBlockAxisIncr[i*9+2];
					int Y = startY + x * cBigBlockAxisIncr[i*9+3] + y * cBigBlockAxisIncr[i*9+4] + z * cBigBlockAxisIncr[i*9+5];
					int Z = startZ + x * cBigBlockAxisIncr[i*9+6] + y * cBigBlockAxisIncr[i*9+7] + z * cBigBlockAxisIncr[i*9+8];

					if (b.Get(x,y,z))
						o[Z][Y][X] = cEmptyGridPos;
				}
	}

private:
	BlockID o[cGridSize][cGridSize][cGridSize]; // indexed by [z][y][x]
};

struct SearchState
{
	// Indicates which block position are in use.
	BlockID usedBigPos[nBigBlocks];
	BlockID usedSmallPos[nSmallBlocks];

	// The currently selected orientations for all blocks.
	int bigBlockOrient[nBigBlocks];
	int smallBlockOrient[nSmallBlocks];

	void Reset()
	{
		for(int i = 0; i < nBigBlocks; ++i)
		{
			usedBigPos[i] = cEmptyGridPos;
			bigBlockOrient[i] = 0;
		}
		for(int i = 0; i < nSmallBlocks; ++i)
		{
			usedSmallPos[i] = cEmptyGridPos;
			smallBlockOrient[i] = 0;
		}
	}
};

/** Search the tree by recursion.
	@param s Stores the current search state. 
	@param g For collision checking.
	@param level Current search level, starts at 0.
	@param p For stats, current completion percentage.
	@param frac For stats, the significance of this search subtree on overall percentage.
	@return True on success, false otherwise. */
bool RecursiveBackTrack(SearchState &s, BlockGrid &g, int level, float p, float frac)
{
	// At levels [0, nBigBlocks[ we plot big blocks on the grid.
	if (level >= 0 && level < nBigBlocks)
	{
		// Calculate stats for the percentage counter
		int nTotalChoices = 0;
		for(int i = 0; i < nBigBlocks; ++i)
			if (s.usedBigPos[i] == cEmptyGridPos)
				++nTotalChoices;
		nTotalChoices *= prBig[level].blocks.size();

		// At a single level of search, try all possible position+orientation pairs for the current piece.
		int curChoice = 0;
		for(int i = 0; i < nBigBlocks; ++i) // Trying all possible positions
		{
			if (s.usedBigPos[i] != cEmptyGridPos) // If this pos is free, we try to plot here, if not, continue to next.
				continue;

			s.usedBigPos[i] = level;
			for(int j = 0; j < prBig[level].blocks.size(); ++j) // Try each orientation at this position.
			{
				// Percentage counter
				float p2 = p + frac*curChoice++/nTotalChoices;
				static int c = 0;
				if (c++ >= 2000000)
				{
					PRINT("p%% done: %f", p2*100.f);
					c = 0;
				}

				if (!g.TryPlaceBlock(prBig[level].blocks[j], i, false, level))
					continue; // Try next orientation, or next position if all orientations have been checked.
				// The piece fit in position i and orientation j, recurse to try out all choices for the next blocks.
				if (RecursiveBackTrack(s, g, level+1, p2, frac/nTotalChoices))
				{
					s.bigBlockOrient[level] = j;
					return true; // We have a global solution, return true through whole recursive call stack.
				}
				// The rest of the blocks could not be placed properly, erase this and try next orientation and/or position.
				g.ClearBlock(prBig[level].blocks[j], i, false);
			}
			// We try the next position, so mark this position free as it was not the correct one.
			s.usedBigPos[i] = cEmptyGridPos;
		}
		// All pos/orientation pairs tested, none worked, return false so the caller can prune the tree.
		return false;
	}
	// At levels [nBigBlocks, nBlocks[ we place small blocks.
	else if (level >= nBigBlocks && level < nBlocks)
	{
		for(int i = 0; i < nSmallBlocks; ++i)
		{
			if (s.usedSmallPos[i] != cEmptyGridPos) // If this pos is taken, continue to next.
				continue;

			s.usedSmallPos[i] = level;
			for(int j = 0; j < prSmall[level-nBigBlocks].blocks.size(); ++j) // Try all orientations for this block.
			{
				if (!g.TryPlaceBlock(prSmall[level-nBigBlocks].blocks[j], i, true, level))
					continue;

				if (RecursiveBackTrack(s, g, level+1, p, frac))
				{
					s.smallBlockOrient[level-nBigBlocks] = j;
					return true;
				}

				g.ClearBlock(prSmall[level-nBigBlocks].blocks[j], i, true);
			}
			s.usedSmallPos[i] = cEmptyGridPos;
		}
		return false;
	}
	else if (level == nBlocks)
	{
		PRINT("Found solution!");
		for(int i = 0; i < nBigBlocks; ++i)
			PRINT("Big pos %d: %d, orientation: %d", i, s.usedBigPos[i], s.bigBlockOrient[s.usedBigPos[i]]);
		for(int i = 0; i < nSmallBlocks; ++i)
			PRINT("Small pos %d: %d, orientation: %d", i, s.usedSmallPos[i], s.smallBlockOrient[s.usedSmallPos[i]]);
		return true;
	}
	assert(false);
	return false;
}

class BurrApp : public clb::ClbLibApp<BurrApp>
{
private:
	BlockGrid grid;

	Camera cam;

	// Resources for rendering the blocks.
	Ptr(Geometry) box;
	Ptr(Effect) fx;
	Ptr(Material) mat;
	Ptr(Texture) labelTextures[nBlocks];
	Ptr(Texture) woodTextures[2];

	// Selects which blocks to render.
	bool enabled[nBlocks];

public:
	void InitProblem()
	{
		// Initializes the pieces to the correct configuration for visualization. 
		const int bigBlockPermutation[nBigBlocks] = { 0,11,2,3,8,5,7,9,6,1,4,10 };
		const int bigBlockOrientation[nBigBlocks] = { 1,7,0,3,5,6,3,0,0,2,1,2 };
		for(int i = 0; i < nBigBlocks; ++i)
		{
			bool success = grid.TryPlaceBlock(prBig[i].blocks[bigBlockOrientation[i]], bigBlockPermutation[i], false, i);
			assert(success);
		}

		const int smallBlockPermutation[nSmallBlocks] = { 0,3,2,5,4,1 };
		const int smallBlockOrientation[nSmallBlocks] = { 0,3,3,1,6,1 };
		for(int i = 0; i < nSmallBlocks; ++i)
		{
			bool success = grid.TryPlaceBlock(prSmall[i].blocks[smallBlockOrientation[i]], smallBlockPermutation[i], true, i + nBigBlocks);
			assert(success);
		}
	}

	void Init()
	{
		grid.Clear();
		for(int i = 0; i < nBlocks; ++i)
			enabled[i] = true;

		cam = Camera(0.1f, 200.f, pi / 4.f, graphics.Width()/(float)graphics.Height());
		cam.SetTargetPos(float3(0,0,0));
		cam.SetUp(float3(0,1,0));
		cam.SetPos(float3(0,0,-30));

		box = clb::GeometryFactory::CreateAABB(0, "box", AABB(float3(0,0,0), float3(1,1,1)), 
			GeometryFactory::TriWindCCW);

		fx = resPool.Create<Effect>("BlockFx");
		mat = resPool.CreateEmpty<Material>("BlockMat");
		for(int i = 0; i < nBlocks; ++i)
		{
			const char labelNames[] = "1234567890QERTYUIO";
			char str[256];
			sprintf(str, "Label%c.png", labelNames[i]);
			labelTextures[i] = resPool.Create<Texture>(str);
		}
		woodTextures[0] = resPool.Create<Texture>("WoodLight.png");
		woodTextures[1] = resPool.Create<Texture>("WoodDark.png");

		InitProblem();
	}

	void FrameMove(float fTimePassed)
	{
		const int key[nBlocks] =
		{
			DIK_1, DIK_2, DIK_3, DIK_4, DIK_5, DIK_6, DIK_7, DIK_8, DIK_9, DIK_0, DIK_Q, DIK_E, DIK_R, DIK_T,
			DIK_Y, DIK_U, DIK_I, DIK_O
		};
		for(int i = 0; i < nBlocks; ++i) // Toggle visibility
			if (input.IsKeyPressed(key[i]))
				enabled[i] = !enabled[i];

		if (input.IsKeyPressed(DIK_P)) // Show all
			for(int i = 0; i < nBlocks; ++i)
				enabled[i] = true;

		if (input.IsKeyPressed(DIK_L)) // Hide all
			for(int i = 0; i < nBlocks; ++i)
				enabled[i] = false;

		const float cMouseS = 5.f * fTimePassed;

		UpdateFreeLook(cam,
			cMouseS * ((int)(input.IsKeyDown(DIK_W)!=0) - (int)(input.IsKeyDown(DIK_S)!=0)),
			cMouseS * ((int)(input.IsKeyDown(DIK_SPACE)!=0) - (int)(input.IsKeyDown(DIK_LCONTROL)!=0)),
			cMouseS * ((int)(input.IsKeyDown(DIK_D)!=0) - (int)(input.IsKeyDown(DIK_A)!=0)),
			input.Get(MouseRelX) * 0.001f, input.Get(MouseRelY) * 0.001f);

	}

	void RenderCube(const float3 &pos, unsigned int cubeNum)
	{
		if (cubeNum == cEmptyGridPos || !enabled[cubeNum])
			return;

		const float3 colors[] =
		{
			float3(0.33f,0,0),
			float3(0.66f,0,0),
			float3(1,0,0),
			float3(0,0.33f,0),
			float3(0,0.66f,0),
			float3(0,1,0),
			float3(0,0,0.33f),
			float3(0,0,0.66f),
			float3(0,0,1),
			float3(0.33f,0,0.33f),
			float3(0.66f,0,0.66f),
			float3(1,0,1),
			float3(0.33f,0.33f,0),
			float3(0.66f,0.66f,0),
			float3(1,1,0),
			float3(0,0.33f,0.33f),
			float3(0,0.66f,0.66f),
			float3(0,1,1)
		};

		mat->Set<float3>("cDiffuse", colors[cubeNum]);
		mat->Set<Ptr(BaseTexture)>("tBase", woodTextures[cubeNum<nBigBlocks?0:1]);
		mat->Set<Ptr(BaseTexture)>("tBase2", labelTextures[cubeNum]);
		float3x4 tm = MatIdentity3x4();
		SetTranslatePart(tm, pos);
		graphics.SetWorldTransform(tm);
		graphics.RenderGeometry(*box, *fx, mat);
	}

	void RenderCube2x2x2(const float3 &pos, unsigned int cubeNum)
	{
		for(int z = 0; z < 2; ++z)
			for(int y = 0; y < 2; ++y)
				for(int x = 0; x < 2; ++x)
					RenderCube(pos + float3(x,y,z), cubeNum);
	}

	void Render()
	{
		graphics.Clear(0x80808080, 1.f);
		graphics.SetViewTransform(cam.ViewMatrix());
		graphics.SetProjTransform(cam.ProjectionMatrix());

		Light l;
		l.bActive = true;
		l.bCastsShadows = false;
		l.bSpecular = false;
		l.brightness = 30.f;
		l.camera = Camera(1.f, 100.f, pi/2.f, 1.f);
		l.camera.SetPos(cam.Pos());
		l.camera.SetUp(float3(0,1,0));
		l.camera.SetTargetPos(float3(0,0,0));
		l.type = LTPoint;
		l.vAmbientColor = float3(0.5f,0.5f,0.5f);
		l.vDiffuseColor = float3(0.4f,0.4f,0.4f);
		graphics.SetLight(0, &l);

		// Render the cubes on the grid
		for(int z = 0; z < 8; ++z)
			for(int y = 0; y < 8; ++y)
				for(int x = 0; x < 8; ++x)
					RenderCube(float3(x,y,z), grid.Get(x,y,z));

		// Now extend all the block ends outside the grid so the puzzle resembles the original.
		RenderCube2x2x2(float3(2,2,-2), grid.Get(2,2,0));
		RenderCube2x2x2(float3(2,2,8),  grid.Get(2,2,0));
		RenderCube2x2x2(float3(4,2,-2), grid.Get(5,2,0));
		RenderCube2x2x2(float3(4,2,8),  grid.Get(5,2,0));
		RenderCube2x2x2(float3(2,4,-2), grid.Get(2,5,0));
		RenderCube2x2x2(float3(2,4,8),  grid.Get(2,5,0));
		RenderCube2x2x2(float3(4,4,-2), grid.Get(5,5,0));
		RenderCube2x2x2(float3(4,4,8),  grid.Get(5,5,0));

		RenderCube2x2x2(float3(2,-2,2), grid.Get(2,0,2));
		RenderCube2x2x2(float3(2,8,2),  grid.Get(2,0,2));
		RenderCube2x2x2(float3(4,-2,2), grid.Get(5,0,2));
		RenderCube2x2x2(float3(4,8,2),  grid.Get(5,0,2));
		RenderCube2x2x2(float3(2,-2,4), grid.Get(2,0,5));
		RenderCube2x2x2(float3(2,8,4),  grid.Get(2,0,5));
		RenderCube2x2x2(float3(4,-2,4), grid.Get(5,0,5));
		RenderCube2x2x2(float3(4,8,4),  grid.Get(5,0,5));

		RenderCube2x2x2(float3(-2,2,2), grid.Get(0,2,2));
		RenderCube2x2x2(float3(8,2,2),  grid.Get(0,2,2));
		RenderCube2x2x2(float3(-2,2,4), grid.Get(0,2,5));
		RenderCube2x2x2(float3(8,2,4),  grid.Get(0,2,5));
		RenderCube2x2x2(float3(-2,4,2), grid.Get(0,5,2));
		RenderCube2x2x2(float3(8,4,2),  grid.Get(0,5,2));
		RenderCube2x2x2(float3(-2,4,4), grid.Get(0,5,5));
		RenderCube2x2x2(float3(8,4,4),  grid.Get(0,5,5));

		graphics.RenderText(*resPool.Get<Font>("systemFont"), 5, graphics.Height()-35, "WSAD/Space/Ctrl/Mouse: Move around", 0xFF0000FF);
		graphics.RenderText(*resPool.Get<Font>("systemFont"), 5, graphics.Height()-25, "1234567890QERTYUIO: Toggle piece visibility", 0xFF0000FF);
		graphics.RenderText(*resPool.Get<Font>("systemFont"), 5, graphics.Height()-15, "P/L: Show/Hide all pieces", 0xFF0000FF);
	}

	void RenderToFrameBuffer() {}
	void DeInit() {}
};

int clbMain()
{
	PrecalculateRotations();
#ifdef CONSOLE_APP
	SearchState s;
	s.Reset();
	BlockGrid g;
	g.Clear();
	RecursiveBackTrack(s, g, 0, 0, 1.f);
#else
	BurrApp app;
	app.Run();
#endif
	return 0;
}
